/*
** Copyright (C) 2004-2013 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.227.7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract FA8721-05-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
**  rwpmaplookup.c
**
**    Read the prefix map file provided on the command line (which may
**    indicate rwpmaplookup should read the default country code or
**    addrtype mapping files).  Then read textual input from files or
**    from command line arguments, parse the text as either IP
**    addresses or as protocol/port pairs depending on the type of
**    prefix map specified, lookup the IP or proto/port pair in the
**    prefix map, and print the results as textual columns.
**
**    When --no-files is specified, the command line arguments are
**    treated as the text to parse and look-up.  Otherwise, read input
**    from the named files, from standard input, or from the files
**    specified by --xargs.
**
**  Mark Thomas
**  June 2011
**
*/

#include <silk/silk.h>

RCSIDENT("$SiLK: rwpmaplookup.c 9f366e88c8bd 2013-06-20 18:29:15Z mthomas $");

#include <silk/skcountry.h>
#include <silk/skipaddr.h>
#include <silk/skipset.h>
#include <silk/skprefixmap.h>
#include <silk/skstream.h>
#include <silk/skstringmap.h>
#include <silk/utils.h>


/* LOCAL DEFINES AND TYPEDEFS */

/* where to write --help output */
#define USAGE_FH stdout

/* max expected length of a pmap dictionary entry */
#define DICTIONARY_ENTRY_BUFLEN 2048

/* value indicating an unset IP address */
#define ADDRESS_UNSET 0xFFFFFFFF

/* maximum number of output fields */
#define PMAP_FIND_MAX_FIELD_COUNT  8

/* for columnar output, the minimum width for the value field */
#define PMAP_FIND_VALUE_WIDTH_MINIMUM   5

/* for columnar output, the default width for the value field when no
 * other value can be determined */
#define PMAP_FIND_VALUE_WIDTH_DEFAULT  20


/* identifiers for the type of prefix map */
typedef enum pmap_type_en {
    PMAP_FIND_TYPE_NONE, PMAP_FIND_TYPE_PROTOPORT,
    PMAP_FIND_TYPE_IPV4, PMAP_FIND_TYPE_IPV6,
    PMAP_FIND_TYPE_COUNTRY, PMAP_FIND_TYPE_ADDRTYPE
} pmap_type_t;

/* identifiers of the fields to print */
typedef enum pmap_find_field_en {
    PMAP_FIND_FIELD_KEY, PMAP_FIND_FIELD_VALUE,
    PMAP_FIND_FIELD_INPUT, PMAP_FIND_FIELD_RANGE,
    PMAP_FIND_FIELD_START, PMAP_FIND_FIELD_END
} pmap_find_field_t;

/* default output fields */
#define PMAP_FIND_DEFAULT_FIELDS  "key,value"


/* LOCAL VARIABLES */

/* string map used to get field names from user */
static const sk_stringmap_entry_t field_map_entries[] = {
    {"key",          PMAP_FIND_FIELD_KEY,                 NULL, NULL},
    {"value",        PMAP_FIND_FIELD_VALUE,               NULL, NULL},
    {"input",        PMAP_FIND_FIELD_INPUT,               NULL, NULL},
/*
    {"range",        PMAP_FIND_FIELD_RANGE,               NULL, NULL},
    {"start",        PMAP_FIND_FIELD_START,               NULL, NULL},
    {"end",          PMAP_FIND_FIELD_END,                 NULL, NULL},
*/
    SK_STRINGMAP_SENTINEL
};

/* fields to print, in the order in which to print them */
static pmap_find_field_t fields[PMAP_FIND_MAX_FIELD_COUNT];

/* number of fields in list */
static size_t num_fields;

/* width of the columns, where index is an pmap_find_field_t. */
static int col_width[PMAP_FIND_MAX_FIELD_COUNT];

/* the type of prefix map */
static pmap_type_t pmap_type = PMAP_FIND_TYPE_NONE;

/* the prefix map file */
static skPrefixMap_t *map = NULL;

/* whether 'fields[]' contains the PMAP_FIND_FIELD_INPUT field */
static int printing_input = 0;

/* when non-zero, do not print errors.  set by --no-errors */
static int no_errors = 0;

/* when non-zero, treat arguments as values to look-up.  set by --no-files */
static int no_files = 0;

/* when non-zero, treat arguments as names of IPset files.  set by
 * --ipset-files */
static int ipset_files = 0;

/* the output stream, set by --output-path or --pager */
static sk_fileptr_t output;

/* paging program */
static const char *pager;

/* flags used to print IPs */
static uint32_t ip_format = SKIPADDR_CANONICAL;

/* delimiter */
static char column_separator = '|';

/* how to format output */
static int no_columns;
static int no_final_delimiter;
static int no_titles;

/* final delimiter */
static char final_delim[] = {'\0', '\0'};

/* for processing input */
static sk_options_ctx_t *optctx;


/* OPTIONS SETUP */

typedef enum {
    OPT_MAP_FILE, OPT_ADDRESS_TYPES, OPT_COUNTRY_CODES,
    OPT_FIELDS,
#if 0
    OPT_INPUT_FIELDS, OPT_INPUT_DELIMITER,
#endif
    OPT_NO_FILES, OPT_NO_ERRORS, OPT_IPSET_FILES,
    OPT_NO_TITLES,
    OPT_NO_COLUMNS, OPT_COLUMN_SEPARATOR, OPT_NO_FINAL_DELIMITER,
    OPT_DELIMITED,
    OPT_OUTPUT_PATH, OPT_PAGER
} appOptionsEnum;

static struct option appOptions[] = {
    {"map-file",            REQUIRED_ARG, 0, OPT_MAP_FILE},
    {"address-types",       OPTIONAL_ARG, 0, OPT_ADDRESS_TYPES},
    {"country-codes",       OPTIONAL_ARG, 0, OPT_COUNTRY_CODES},
    {"fields",              REQUIRED_ARG, 0, OPT_FIELDS},
#if 0
    {"input-fields",        REQUIRED_ARG, 0, OPT_INPUT_FIELDS},
    {"input-delimiter",     REQUIRED_ARG, 0, OPT_INPUT_DELIMITER},
#endif
    {"no-files",            NO_ARG,       0, OPT_NO_FILES},
    {"no-errors",           NO_ARG,       0, OPT_NO_ERRORS},
    {"ipset-files",         NO_ARG,       0, OPT_IPSET_FILES},

    {"no-titles",           NO_ARG,       0, OPT_NO_TITLES},
    {"no-columns",          NO_ARG,       0, OPT_NO_COLUMNS},
    {"column-separator",    REQUIRED_ARG, 0, OPT_COLUMN_SEPARATOR},
    {"no-final-delimiter",  NO_ARG,       0, OPT_NO_FINAL_DELIMITER},
    {"delimited",           OPTIONAL_ARG, 0, OPT_DELIMITED},
    {"output-path",         REQUIRED_ARG, 0, OPT_OUTPUT_PATH},
    {"pager",               REQUIRED_ARG, 0, OPT_PAGER},
    {0,0,0,0}               /* sentinel entry */
};

static const char *appHelp[] = {
    ("Find the IPs or the protocol/port pairs in the named\n"
     "\t prefix map file"),
    ("Find IPs in the named address types mapping file, or\n"
     "\tin the default file when no argument is provided"),
    ("Find IPs in the named country code mapping file, or\n"
     "\tin the default file when no argument is provided"),
    NULL,  /* generated dynamically */
#if 0
    "Names of input columns",
    "Character separating input columns",
#endif  /* 0 */
    ("Do not read from files and instead treat the command\n"
     "\tline arguments as the IPs or proto/port pairs to find. Def. No"),
    "Do not report errors parsing the input. Def. No",
    ("Treat the command line arguments as binary IPset files to\n"
     "\tread. Def. Treat command line arguments as names of text files"),

    "Do not print column headers. Def. Print titles.",
    "Disable fixed-width columnar output. Def. Columnar",
    "Use specified character between columns. Def. '|'",
    "Suppress column delimiter at end of line. Def. No",
    "Shortcut for --no-columns --no-final-del --column-sep=CHAR",
    "Send output to given file path. Def. stdout",
    "Program to invoke to page output. Def. $SILK_PAGER or $PAGER",
    (char *)NULL
};


/* LOCAL FUNCTION PROTOTYPES */

static int  appOptionsHandler(clientData cData, int opt_index, char *opt_arg);
static void usageFields(FILE *fh);
static int  parseFields(const char *field_string);


/* FUNCTION DEFINITIONS */

/*
 *  appUsageLong();
 *
 *    Print complete usage information to USAGE_FH.  Pass this
 *    function to skOptionsSetUsageCallback(); skOptionsParse() will
 *    call this funciton and then exit the program when the --help
 *    option is given.
 */
static void appUsageLong(void)
{
#define USAGE_MSG                                                             \
    ("<SWITCHES>\n"                                                           \
     "\tMap textual IPs, textual protocol/port pairs, or addresses in a\n"    \
     "\tbinary IPset file to entries in a binary prefix map and print the\n"  \
     "\tresults in a columnar, |-delimited format.\n"                         \
     "\tExactly one of --map-file, --address-types, or --country-codes must\n" \
     "\tbe specified, where --map-file requires a prefix map argument and\n"  \
     "\tthe other switches use the default map unless an argument is\n"       \
     "\tprovided to the switch.  The textual input is read from files\n"      \
     "\tspecified on the command line, or you may specify the --no-files\n"   \
     "\tswitch and specify the IP(s) or protocol/port pair(s) on the\n"       \
     "\tcommand line.  Use --ipset-files to process data from IPset files.\n")

    FILE *fh = USAGE_FH;
    int i;

    fprintf(fh, "%s %s", skAppName(), USAGE_MSG);
    fprintf(fh, "\nSWITCHES:\n");
    skOptionsDefaultUsage(fh);

    for (i = 0; appOptions[i].name; i++ ) {
        fprintf(fh, "--%s %s. ", appOptions[i].name,
                SK_OPTION_HAS_ARG(appOptions[i]));
        switch (appOptions[i].val) {
          case OPT_FIELDS:
            /* Dynamically build the help */
            usageFields(fh);
            break;
          case OPT_IPSET_FILES:
            fprintf(fh, "%s\n", appHelp[i]);
            /* insert the --ip-format switches */
            skOptionsIPFormatUsage(fh);
            break;
          default:
            /* Simple static help text from the appHelp array */
            fprintf(fh, "%s\n", appHelp[i]);
            break;
        }
    }

    skOptionsCtxOptionsUsage(optctx, fh);
}


/*
 *  appTeardown()
 *
 *    Teardown all modules, close all files, and tidy up all
 *    application state.
 *
 *    This function is idempotent.
 */
static void appTeardown(void)
{
    static int teardownFlag = 0;

    if (teardownFlag) {
        return;
    }
    teardownFlag = 1;

    /* close the output file or process */
    if (output.of_name) {
        skFileptrClose(&output, &skAppPrintErr);
    }

    switch (pmap_type) {
      case PMAP_FIND_TYPE_COUNTRY:
        skCountryTeardown();
        break;
      case PMAP_FIND_TYPE_ADDRTYPE:
        skAddressTypesTeardown();
        break;
      case PMAP_FIND_TYPE_IPV4:
      case PMAP_FIND_TYPE_IPV6:
      case PMAP_FIND_TYPE_PROTOPORT:
        if (map) {
            skPrefixMapDelete(map);
        }
        break;
      case PMAP_FIND_TYPE_NONE:
        break;
    }
    map = NULL;

    skOptionsCtxDestroy(&optctx);
    skAppUnregister();
}


/*
 *  appSetup(argc, argv);
 *
 *    Perform all the setup for this application include setting up
 *    required modules, parsing options, etc.  This function should be
 *    passed the same arguments that were passed into main().
 *
 *    Returns to the caller if all setup succeeds.  If anything fails,
 *    this function will cause the application to exit with a FAILURE
 *    exit status.
 */
static void appSetup(int argc, char **argv)
{
    SILK_FEATURES_DEFINE_STRUCT(features);
    int optctx_flags;
    int rv;

    /* verify same number of options and help strings */
    assert((sizeof(appHelp)/sizeof(char *)) ==
           (sizeof(appOptions)/sizeof(struct option)));

    /* register the application */
    skAppRegister(argv[0]);
    skAppVerifyFeatures(&features, NULL);
    skOptionsSetUsageCallback(&appUsageLong);

    /* initialize globals */
    memset(&output, 0, sizeof(output));
    output.of_fp = stdout;

    optctx_flags = (SK_OPTIONS_CTX_ALLOW_STDIN | SK_OPTIONS_CTX_XARGS);

    /* register the options */
    if (skOptionsCtxCreate(&optctx, optctx_flags)
        || skOptionsCtxOptionsRegister(optctx)
        || skOptionsRegister(appOptions, &appOptionsHandler, NULL)
        || skOptionsIPFormatRegister(&ip_format))
    {
        skAppPrintErr("Unable to register options");
        exit(EXIT_FAILURE);
    }

    /* register the teardown hanlder */
    if (atexit(appTeardown) < 0) {
        skAppPrintErr("Unable to register appTeardown() with atexit()");
        appTeardown();
        exit(EXIT_FAILURE);
    }

    /* parse the options */
    rv = skOptionsCtxOptionsParse(optctx, argc, argv);
    if (rv < 0) {
        skAppUsage();           /* never returns */
    }

    /* ensure we have a pmap */
    if (PMAP_FIND_TYPE_NONE == pmap_type) {
        skAppPrintErr("Must specify the prefix map to use");
        exit(EXIT_FAILURE);
    }

    /* do not query a protocol/port map for IPs */
    if (PMAP_FIND_TYPE_PROTOPORT == pmap_type && ipset_files) {
        skAppPrintErr("May not use a %s prefix map with --%s",
                      skPrefixMapGetContentName(pmap_type),
                      appOptions[OPT_IPSET_FILES].name);
        exit(EXIT_FAILURE);
    }

    /* when --no-files is specified, ensure we have arguments on the
     * command line */
    if (no_files && 0 == skOptionsCtxCountArgs(optctx)) {
        skAppPrintErr("Must provide command line arguments when --%s is used",
                      appOptions[OPT_NO_FILES].name);
        exit(EXIT_FAILURE);
    }

    /* set default fields */
    if (0 == num_fields) {
        rv = parseFields(PMAP_FIND_DEFAULT_FIELDS);
        assert(0 == rv);
    }

    /* set variables for formatting the output */
    if (!no_final_delimiter) {
        final_delim[0] = column_separator;
    }
    memset(col_width, 0, sizeof(col_width));
    if (!no_columns) {
        col_width[PMAP_FIND_FIELD_INPUT] = PMAP_FIND_VALUE_WIDTH_DEFAULT;
        col_width[PMAP_FIND_FIELD_VALUE] = PMAP_FIND_VALUE_WIDTH_DEFAULT;

        switch (pmap_type) {
          case PMAP_FIND_TYPE_COUNTRY:
            col_width[PMAP_FIND_FIELD_VALUE] = PMAP_FIND_VALUE_WIDTH_MINIMUM;
            break;
          case PMAP_FIND_TYPE_ADDRTYPE:
          case PMAP_FIND_TYPE_IPV4:
          case PMAP_FIND_TYPE_IPV6:
          case PMAP_FIND_TYPE_PROTOPORT:
            if (skPrefixMapDictionaryGetWordCount(map) > 0) {
                col_width[PMAP_FIND_FIELD_VALUE]
                    = (int)skPrefixMapDictionaryGetMaxWordSize(map);
                if (col_width[PMAP_FIND_FIELD_VALUE]
                    < PMAP_FIND_VALUE_WIDTH_MINIMUM)
                {
                    col_width[PMAP_FIND_FIELD_VALUE]
                        = PMAP_FIND_VALUE_WIDTH_MINIMUM;
                }
            }
            break;
          case PMAP_FIND_TYPE_NONE:
            skAbortBadCase(pmap_type);
        }

        if (PMAP_FIND_TYPE_PROTOPORT == pmap_type) {
            col_width[PMAP_FIND_FIELD_KEY] = 3 + 1 + 5;
        } else if (PMAP_FIND_TYPE_IPV6 == pmap_type) {
            if (ip_format == SKIPADDR_HEXADECIMAL) {
                col_width[PMAP_FIND_FIELD_KEY] = 32;
            } else {
                col_width[PMAP_FIND_FIELD_KEY] = 39;
            }
            col_width[PMAP_FIND_FIELD_RANGE]
                = col_width[PMAP_FIND_FIELD_KEY] + 1 + 3;
        } else {
            if (ip_format == SKIPADDR_DECIMAL) {
                col_width[PMAP_FIND_FIELD_KEY] = 10;
            } else if (ip_format == SKIPADDR_HEXADECIMAL) {
                col_width[PMAP_FIND_FIELD_KEY] = 8;
            } else {
                col_width[PMAP_FIND_FIELD_KEY] = 15;
            }
            col_width[PMAP_FIND_FIELD_RANGE]
                = col_width[PMAP_FIND_FIELD_KEY] + 1 + 2;
        }
        col_width[PMAP_FIND_FIELD_START]
            = col_width[PMAP_FIND_FIELD_END]
            = col_width[PMAP_FIND_FIELD_KEY];
    }

    /* create the output stream. */
    /* open the --output-path.  the 'of_name' member is NULL if user
     * didn't get an output-path.  do not page when no_files is active
     * or when an explicit output-path is given */
    if (output.of_name) {
        rv = skFileptrOpen(&output, SK_IO_WRITE);
        if (rv) {
            skAppPrintErr("Unable to open %s '%s': %s",
                          appOptions[OPT_OUTPUT_PATH].name,
                          output.of_name, skFileptrStrerror(rv));
            exit(EXIT_FAILURE);
        }
    } else if (!no_files) {
        /* Invoke the pager */
        rv = skFileptrOpenPager(&output, pager);
        if (rv && rv != SK_FILEPTR_PAGER_IGNORED) {
            skAppPrintErr("Unable to invoke pager");
        }
    }

    return;  /* OK */
}


/*
 *  status = appOptionsHandler(cData, opt_index, opt_arg);
 *
 *    This function is passed to skOptionsRegister(); it will be called
 *    by skOptionsParse() for each user-specified switch that the
 *    application has registered; it should handle the switch as
 *    required---typically by setting global variables---and return 1
 *    if the switch processing failed or 0 if it succeeded.  Returning
 *    a non-zero from from the handler causes skOptionsParse() to return
 *    a negative value.
 *
 *    The clientData in 'cData' is typically ignored; 'opt_index' is
 *    the index number that was specified as the last value for each
 *    struct option in appOptions[]; 'opt_arg' is the user's argument
 *    to the switch for options that have a REQUIRED_ARG or an
 *    OPTIONAL_ARG.
 */
static int appOptionsHandler(
    clientData  UNUSED(cData),
    int         opt_index,
    char       *opt_arg)
{
    int rv;

    switch ((appOptionsEnum)opt_index) {
      case OPT_MAP_FILE:
        if (pmap_type != PMAP_FIND_TYPE_NONE) {
            skAppPrintErr("Invalid %s: May only specify one prefix map file",
                          appOptions[opt_index].name);
            return -1;
        }
        rv = skPrefixMapLoad(&map, opt_arg);
        if (rv) {
            return -1;
        }
        switch (skPrefixMapGetContentType(map)) {
          case SKPREFIXMAP_CONT_PROTO_PORT:
            pmap_type = PMAP_FIND_TYPE_PROTOPORT;
            break;
          case SKPREFIXMAP_CONT_ADDR_V4:
            pmap_type = PMAP_FIND_TYPE_IPV4;
            break;
          case SKPREFIXMAP_CONT_ADDR_V6:
            pmap_type = PMAP_FIND_TYPE_IPV6;
            break;
          default:
            skAbortBadCase(skPrefixMapGetContentType(map));
        }
        break;

      case OPT_COUNTRY_CODES:
        if (pmap_type != PMAP_FIND_TYPE_NONE) {
            skAppPrintErr("Invalid %s: May only specify one prefix map file",
                          appOptions[opt_index].name);
            return -1;
        }
        /* find and load the map file */
        if (skCountrySetup(opt_arg, &skAppPrintErr)) {
            return -1;
        }
        pmap_type = PMAP_FIND_TYPE_COUNTRY;
        break;

      case OPT_ADDRESS_TYPES:
        if (pmap_type != PMAP_FIND_TYPE_NONE) {
            skAppPrintErr("Invalid %s: May only specify one prefix map file",
                          appOptions[opt_index].name);
            return -1;
        }
        /* find and load the map file */
        if (skAddressTypesSetup(opt_arg, &skAppPrintErr)) {
            return -1;
        }
        map = skAddressTypesGetPmap();
        pmap_type = PMAP_FIND_TYPE_ADDRTYPE;
        break;

      case OPT_FIELDS:
        if (parseFields(opt_arg)) {
            return -1;
        }
        break;

#if 0
      case OPT_INPUT_FIELDS:
        break;

      case OPT_INPUT_DELIMITER:
        break;
#endif  /* 0 */

      case OPT_IPSET_FILES:
        if (no_files) {
            skAppPrintErr("Invalid %s: May not be combined with --%s",
                          appOptions[opt_index].name,
                          appOptions[OPT_NO_FILES].name);
            return -1;
        }
        ipset_files = 1;
        break;

      case OPT_NO_FILES:
        if (ipset_files) {
            skAppPrintErr("Invalid %s: May not be combined with --%s",
                          appOptions[opt_index].name,
                          appOptions[OPT_IPSET_FILES].name);
            return -1;
        }
        no_files = 1;
        break;

      case OPT_NO_ERRORS:
        no_errors = 1;
        break;

      case OPT_NO_TITLES:
        no_titles = 1;
        break;

      case OPT_NO_COLUMNS:
        no_columns = 1;
        break;

      case OPT_COLUMN_SEPARATOR:
        column_separator = opt_arg[0];
        break;

      case OPT_NO_FINAL_DELIMITER:
        no_final_delimiter = 1;
        break;

      case OPT_DELIMITED:
        no_columns = 1;
        no_final_delimiter = 1;
        if (opt_arg) {
            column_separator = opt_arg[0];
        }
        break;

      case OPT_OUTPUT_PATH:
        if (output.of_name) {
            skAppPrintErr("Invalid %s: Switch used multiple times",
                          appOptions[opt_index].name);
            return 1;
        }
        output.of_name = opt_arg;
        break;

      case OPT_PAGER:
        pager = opt_arg;
        break;
    }

    return 0;  /* OK */
}


/*
 *  stringmap = createStringmap();
 *
 *    Create the string map that is used to parse the --fields
 *    paramater.
 */
static sk_stringmap_t *createStringmap(void)
{
    sk_stringmap_t *field_map;

    /* Create the map if necessary */
    if (SKSTRINGMAP_OK != skStringMapCreate(&field_map)) {
        return NULL;
    }

    /* add entries */
    if (skStringMapAddEntries(field_map, -1, field_map_entries)
        != SKSTRINGMAP_OK)
    {
        skStringMapDestroy(field_map);
        return NULL;
    }

    return field_map;
}


/*
 *  usageFields(fh);
 *
 *    Print the usage (help) message for --fields to the 'fh' file
 *    pointer.
 */
static void usageFields(FILE *fh)
{
    sk_stringmap_t *field_map;

    /* Create the string map for --fields */
    field_map = createStringmap();
    if (NULL == field_map) {
        fprintf(fh, "Field(s) to print.\n");
        return;
    }

    fprintf(fh, ("Fields(s) to print. Def. " PMAP_FIND_DEFAULT_FIELDS "\n"
                 "\tList field names or IDs separated by commas."
                 "Supported fields:\n"));
    skStringMapPrintUsage(field_map, fh, 8);
    skStringMapDestroy(field_map);
}


/*
 *  ok = parseFields(string);
 *
 *    Parse the output fields listed in 'string' and set the global
 *    fields[] array to the pmap_find_field_t IDs.  Return 0 on
 *    success, or non-zero on failure.
 */
static int parseFields(
    const char *field_string)
{
    sk_stringmap_t *field_map = NULL;
    sk_stringmap_iter_t *iter = NULL;
    sk_stringmap_entry_t *entry;
    char *errmsg;
    int rv = -1;

    if (field_string == NULL || field_string[0] == '\0') {
        skAppPrintErr("Invalid %s: Value is empty",
                      appOptions[OPT_FIELDS].name);
        return rv;
    }
    field_map = createStringmap();
    if (NULL == field_map) {
        skAppPrintOutOfMemory(NULL);
        goto END;
    }

    /* parse the field-list */
    if (skStringMapParse(field_map, field_string, SKSTRINGMAP_DUPES_ERROR,
                         &iter, &errmsg))
    {
        skAppPrintErr("Invalid %s: %s",
                      appOptions[OPT_FIELDS].name, errmsg);
        goto END;
    }

    /* add the selected fields to the 'fields[]' array */
    for (num_fields = 0;
         skStringMapIterNext(iter, &entry, NULL) == SK_ITERATOR_OK;
         ++num_fields)
    {
        assert(num_fields < PMAP_FIND_MAX_FIELD_COUNT);
        fields[num_fields] = (pmap_find_field_t)entry->id;
        if (PMAP_FIND_FIELD_INPUT == fields[num_fields]) {
            printing_input = 1;
        }
    }

    rv = 0;

  END:
    if (iter != NULL) {
        skStringMapIterDestroy(iter);
    }
    if (field_map != NULL) {
        skStringMapDestroy(field_map);
    }
    return rv;
}


/*
 *  printTitles();
 *
 *    Print the column headings unless --no-titles was specified.
 */
static void printTitles(void)
{
    size_t i;

    if (no_titles) {
        return;
    }

    for (i = 0; i < num_fields; ++i) {
        if (i > 0) {
            fprintf(output.of_fp, "%c", column_separator);
        }
        fprintf(output.of_fp, "%*s",
                col_width[fields[i]], field_map_entries[fields[i]].name);
    }
    fprintf(output.of_fp, "%s\n", final_delim);
}


/*
 *  status = processAddress(string, filename, linenum);
 *
 *    Parse 'string' as an IP address, find the IP address in the
 *    prefix map, and print the resulting columns specified in
 *    --fields.
 *
 *    If an error occurs in parsing, report an error unless
 *    --no-errors was specified.  The parameters 'filename' and
 *    'linenum' specify the source of 'string' and will be used when
 *    reporting the error.  If 'filename' is NULL, it is assumed the
 *    string was read off the command line.
 */
static int processAddress(
    const char *string,
    const char *filename,
    int         linenum)
{
    char buf[SK_NUM2DOT_STRLEN];
    char label[DICTIONARY_ENTRY_BUFLEN];
    skipaddr_t ip;
    size_t i;
    int not_found = 1;
    int rv;

    /* parse the line */
    rv = skStringParseIP(&ip, string);
    if (rv) {
        if (SKUTILS_ERR_EMPTY == rv) {
            goto PRINT_LINE;
        }
        if (no_errors) {
            /* do nothing */
        } else if (filename) {
            skAppPrintErr("Invalid IP '%s' at %s:%d: %s",
                          string, filename, linenum,
                          skStringParseStrerror(rv));
        } else {
            skAppPrintErr("Invalid IP '%s' on command line: %s",
                          string, skStringParseStrerror(rv));
        }
        goto PRINT_LINE;
    }
#if SK_ENABLE_IPV6
    if (map
        && (skPrefixMapGetContentType(map) == SKPREFIXMAP_CONT_ADDR_V4)
        && skipaddrIsV6(&ip))
    {
        goto PRINT_LINE;
    }
#endif /* SK_ENABLE_IPV6 */

    /* get the label for the IP */
    not_found = 0;
    switch (pmap_type) {
      case PMAP_FIND_TYPE_COUNTRY:
        skCountryLookupName(&ip, label, sizeof(label));
        break;
      case PMAP_FIND_TYPE_ADDRTYPE:
      case PMAP_FIND_TYPE_IPV4:
      case PMAP_FIND_TYPE_IPV6:
        skPrefixMapGetString(map, &ip, label, sizeof(label));
        break;
      case PMAP_FIND_TYPE_PROTOPORT:
      case PMAP_FIND_TYPE_NONE:
        skAbortBadCase(pmap_type);
    }

  PRINT_LINE:
    if (not_found) {
        if (!printing_input) {
            return 0;
        }
        for (i = 0; i < num_fields; ++i) {
            if (i > 0) {
                fprintf(output.of_fp, "%c", column_separator);
            }
            switch (fields[i]) {
              case PMAP_FIND_FIELD_KEY:
              case PMAP_FIND_FIELD_VALUE:
              case PMAP_FIND_FIELD_START:
              case PMAP_FIND_FIELD_END:
              case PMAP_FIND_FIELD_RANGE:
                fprintf(output.of_fp, "%*s", col_width[fields[i]], "");
                break;
              case PMAP_FIND_FIELD_INPUT:
                fprintf(output.of_fp, "%*s", col_width[fields[i]], string);
                break;
            }
        }
        fprintf(output.of_fp, "%s\n", final_delim);
        return 0;
    }

    for (i = 0; i < num_fields; ++i) {
        if (i > 0) {
            fprintf(output.of_fp, "%c", column_separator);
        }
        switch (fields[i]) {
          case PMAP_FIND_FIELD_KEY:
            skipaddrString(buf, &ip, ip_format);
            fprintf(output.of_fp, "%*s", col_width[fields[i]], buf);
            break;

          case PMAP_FIND_FIELD_VALUE:
            fprintf(output.of_fp, "%*s", col_width[fields[i]], label);
            break;

          case PMAP_FIND_FIELD_INPUT:
            fprintf(output.of_fp, "%*s", col_width[fields[i]], string);
            break;

          case PMAP_FIND_FIELD_START:
            skipaddrString(buf, &ip, ip_format);
            fprintf(output.of_fp, "%*s", col_width[fields[i]], buf);
            break;

          case PMAP_FIND_FIELD_END:
            skipaddrString(buf, &ip, ip_format);
            fprintf(output.of_fp, "%*s", col_width[fields[i]], buf);
            break;

          case PMAP_FIND_FIELD_RANGE:
            skipaddrString(buf, &ip, ip_format);
            fprintf(output.of_fp, "%*s", col_width[fields[i]], buf);
            break;
        }
    }
    fprintf(output.of_fp, "%s\n", final_delim);

    return 0;
}


/*
 *  status = processProtoPort(string, filename, linenum);
 *
 *    Parse 'string' as a protocol/port pair, find the pair in the
 *    prefix map, and print the resulting columns specified in
 *    --fields.
 *
 *    If an error occurs in parsing, report an error unless
 *    --no-errors was specified.  The parameters 'filename' and
 *    'linenum' specify the source of 'string' and will be used when
 *    reporting the error.  If 'filename' is NULL, it is assumed the
 *    string was read off the command line.
 */
static int processProtoPort(
    const char *string,
    const char *filename,
    int         linenum)
{
    char buf[128];
    char label[DICTIONARY_ENTRY_BUFLEN];
    skPrefixMapProtoPort_t pp;
    uint32_t proto;
    uint32_t port;
    size_t i;
    int not_found = 1;
    int rv;

    /* protocol */
    rv = skStringParseUint32(&proto, string, 0, 255);
    if (rv < 0) {
        if (SKUTILS_ERR_EMPTY == rv) {
            goto PRINT_LINE;
        }
        if (no_errors) {
            /* do nothing */
        } else if (filename) {
            skAppPrintErr("Invalid protocol '%s' at %s:%d: %s",
                          string, filename, linenum,
                          skStringParseStrerror(rv));
        } else {
            skAppPrintErr("Invalid protocol '%s' on command line: %s",
                          string, skStringParseStrerror(rv));
        }
        goto PRINT_LINE;
    }
    if (rv == 0) {
        if (no_errors) {
            /* do nothing */
        } else if (filename) {
            skAppPrintErr(("Invalid proto/port '%s' at %s:%d:"
                           " Missing '/' delimiter"),
                          string, filename, linenum);
        } else {
            skAppPrintErr(("Invalid proto/port '%s' on command line:"
                           " Missing '/' delimiter"),
                          string);
        }
        rv = -1;
        goto PRINT_LINE;
    }

    /* rv is location of the '/' */
    string += rv + 1;

    /* port */
    rv = skStringParseUint32(&port, string, 0, UINT16_MAX);
    if (rv) {
        if (no_errors) {
        } else if (filename) {
            skAppPrintErr("Invalid port '%s' at %s:%d: %s",
                          string, filename, linenum,
                          skStringParseStrerror(rv));
        } else {
            skAppPrintErr("Invalid port '%s' on command line: %s",
                          string, skStringParseStrerror(rv));
        }
        goto PRINT_LINE;
    }

    /* get the label for the proto/port pair */
    not_found = 0;
    pp.proto = proto;
    pp.port = port;
    skPrefixMapGetString(map, &pp, label, sizeof(label));

  PRINT_LINE:
    if (not_found) {
        if (!printing_input) {
            return 0;
        }
        for (i = 0; i < num_fields; ++i) {
            if (i > 0) {
                fprintf(output.of_fp, "%c", column_separator);
            }
            switch (fields[i]) {
              case PMAP_FIND_FIELD_KEY:
              case PMAP_FIND_FIELD_VALUE:
              case PMAP_FIND_FIELD_START:
              case PMAP_FIND_FIELD_END:
              case PMAP_FIND_FIELD_RANGE:
                fprintf(output.of_fp, "%*s", col_width[fields[i]], "");
                break;
              case PMAP_FIND_FIELD_INPUT:
                fprintf(output.of_fp, "%*s", col_width[fields[i]], string);
                break;
            }
        }
        fprintf(output.of_fp, "%s\n", final_delim);
        return 0;
    }

    for (i = 0; i < num_fields; ++i) {
        if (i > 0) {
            fprintf(output.of_fp, "%c", column_separator);
        }
        switch (fields[i]) {
          case PMAP_FIND_FIELD_KEY:
            snprintf(buf, sizeof(buf), "%u/%u", proto, port);
            fprintf(output.of_fp, "%*s", col_width[fields[i]], buf);
            break;

          case PMAP_FIND_FIELD_VALUE:
            fprintf(output.of_fp, "%*s", col_width[fields[i]], label);
            break;

          case PMAP_FIND_FIELD_INPUT:
            fprintf(output.of_fp, "%*s", col_width[fields[i]], string);
            break;

          case PMAP_FIND_FIELD_START:
            snprintf(buf, sizeof(buf), "%u/%u", proto, port);
            fprintf(output.of_fp, "%*s", col_width[fields[i]], buf);
            break;

          case PMAP_FIND_FIELD_END:
            snprintf(buf, sizeof(buf), "%u/%u", proto, port);
            fprintf(output.of_fp, "%*s", col_width[fields[i]], buf);
            break;

          case PMAP_FIND_FIELD_RANGE:
            skAbortBadCase(fields[i]);
        }
    }
    fprintf(output.of_fp, "%s\n", final_delim);

    return 0;
}


/*
 *  status = processInputFile(filein);
 *
 *    For every line in 'filein', invoke processAddress() or
 *    processProtoPort() as appropriate.
 */
static int processInputFile(
    const char             *f_name)
{
    char line[2048];
    skstream_t *stream = NULL;
    int retval = 1;
    int rv;
    int lc = 0;

    /* open input */
    if ((rv = skStreamCreate(&stream, SK_IO_READ, SK_CONTENT_TEXT))
        || (rv = skStreamBind(stream, f_name))
        || (rv = skStreamSetCommentStart(stream, "#"))
        || (rv = skStreamOpen(stream)))
    {
        skStreamPrintLastErr(stream, rv, &skAppPrintErr);
        goto END;
    }

    if (PMAP_FIND_TYPE_PROTOPORT == pmap_type) {
        /* read until end of file */
        while ((rv = skStreamGetLine(stream, line, sizeof(line), &lc))
               != SKSTREAM_ERR_EOF)
        {
            switch (rv) {
              case SKSTREAM_OK:
                /* good, we got our line */
                break;
              case SKSTREAM_ERR_LONG_LINE:
                /* bad: line was longer than sizeof(line) */
                if (!no_errors) {
                    skAppPrintErr("Input line %s:%d too long. ignored",
                                  f_name, lc);
                }
                continue;
              default:
                /* unexpected error */
                skStreamPrintLastErr(stream, rv, &skAppPrintErr);
                goto END;
            }

            processProtoPort(line, f_name, lc);
        }

    } else {
        while ((rv = skStreamGetLine(stream, line, sizeof(line), &lc))
               != SKSTREAM_ERR_EOF)
        {
            switch (rv) {
              case SKSTREAM_OK:
                /* good, we got our line */
                break;
              case SKSTREAM_ERR_LONG_LINE:
                /* bad: line was longer than sizeof(line) */
                skAppPrintErr("Input line %s:%d too long. ignored",
                              f_name, lc);
                continue;
              default:
                /* unexpected error */
                skStreamPrintLastErr(stream, rv, &skAppPrintErr);
                goto END;
            }

            processAddress(line, f_name, lc);
        }
    }

    retval = 0;

  END:
    skStreamDestroy(&stream);
    return retval;
}


/*
 *  status = processIPSetFile(filein);
 *
 *    For every address in 'filein', invoke processAddress() or
 *    processProtoPort() as appropriate.
 */
static int processIPSetFile(
    const char             *f_name)
{
    char buf[SK_NUM2DOT_STRLEN];
    char label[DICTIONARY_ENTRY_BUFLEN];
    skstream_t *stream = NULL;
    skipset_iterator_t iter;
    skipset_t *ipset;
    skipaddr_t ip;
    uint32_t prefix;
    size_t i;
    int rv;

    /* Read the IPset file from the input */
    if ((rv = skStreamCreate(&stream, SK_IO_READ, SK_CONTENT_SILK))
        || (rv = skStreamBind(stream, f_name))
        || (rv = skStreamOpen(stream)))
    {
        skStreamPrintLastErr(stream, rv, &skAppPrintErr);
        skStreamDestroy(&stream);
        return 1;
    }
    rv = skIPSetRead(&ipset, stream);
    if (rv) {
        if (SKIPSET_ERR_FILEIO == rv) {
            skStreamPrintLastErr(stream, skStreamGetLastReturnValue(stream),
                                 &skAppPrintErr);
        } else {
            skAppPrintErr("Unable to read IPset from '%s': %s",
                          f_name, skIPSetStrerror(rv));
        }
        skStreamDestroy(&stream);
        return 1;
    }
    skStreamDestroy(&stream);

    skIPSetIteratorBind(&iter, ipset, 0, SK_IPV6POLICY_MIX);

    while (skIPSetIteratorNext(&iter, &ip, &prefix) == SK_ITERATOR_OK) {
#if SK_ENABLE_IPV6
        if (map
            && (skPrefixMapGetContentType(map) == SKPREFIXMAP_CONT_ADDR_V4)
            && skipaddrIsV6(&ip))
        {
            if (printing_input) {
                for (i = 0; i < num_fields; ++i) {
                    if (i > 0) {
                        fprintf(output.of_fp, "%c", column_separator);
                    }
                    switch (fields[i]) {
                      case PMAP_FIND_FIELD_KEY:
                      case PMAP_FIND_FIELD_VALUE:
                      case PMAP_FIND_FIELD_START:
                      case PMAP_FIND_FIELD_END:
                      case PMAP_FIND_FIELD_RANGE:
                        fprintf(output.of_fp, "%*s", col_width[fields[i]], "");
                        break;
                      case PMAP_FIND_FIELD_INPUT:
                        skipaddrString(buf, &ip, SKIPADDR_CANONICAL);
                        fprintf(output.of_fp, "%*s",
                                col_width[fields[i]], buf);
                        break;
                    }
                }
                fprintf(output.of_fp, "%s\n", final_delim);
            }
            continue;
        }
#endif  /* SK_ENABLE_IPV6 */

        /* get the label for the IP */
        switch (pmap_type) {
          case PMAP_FIND_TYPE_COUNTRY:
            skCountryLookupName(&ip, label, sizeof(label));
            break;
          case PMAP_FIND_TYPE_ADDRTYPE:
          case PMAP_FIND_TYPE_IPV4:
          case PMAP_FIND_TYPE_IPV6:
            skPrefixMapGetString(map, &ip, label, sizeof(label));
            break;
          case PMAP_FIND_TYPE_PROTOPORT:
          case PMAP_FIND_TYPE_NONE:
            skAbortBadCase(pmap_type);
        }

        for (i = 0; i < num_fields; ++i) {
            if (i > 0) {
                fprintf(output.of_fp, "%c", column_separator);
            }
            switch (fields[i]) {
              case PMAP_FIND_FIELD_KEY:
                skipaddrString(buf, &ip, ip_format);
                fprintf(output.of_fp, "%*s", col_width[fields[i]], buf);
                break;

              case PMAP_FIND_FIELD_VALUE:
                fprintf(output.of_fp, "%*s", col_width[fields[i]], label);
                break;

              case PMAP_FIND_FIELD_INPUT:
                skipaddrString(buf, &ip, SKIPADDR_CANONICAL);
                fprintf(output.of_fp, "%*s", col_width[fields[i]], buf);
                break;

              case PMAP_FIND_FIELD_START:
                skipaddrString(buf, &ip, ip_format);
                fprintf(output.of_fp, "%*s", col_width[fields[i]], buf);
                break;

              case PMAP_FIND_FIELD_END:
                skipaddrString(buf, &ip, ip_format);
                fprintf(output.of_fp, "%*s", col_width[fields[i]], buf);
                break;

              case PMAP_FIND_FIELD_RANGE:
                skipaddrString(buf, &ip, ip_format);
                fprintf(output.of_fp, "%*s", col_width[fields[i]], buf);
                break;
            }
        }
        fprintf(output.of_fp, "%s\n", final_delim);
    }

    skIPSetDestroy(&ipset);

    return 0;
}


int main(int argc, char **argv)
{
    char *arg;
    int rv = 0;

    appSetup(argc, argv);                       /* never returns on error */

    printTitles();
    while (skOptionsCtxNextArgument(optctx, &arg) == 0 && rv == 0) {
        if (ipset_files) {
            rv = processIPSetFile(arg);
        } else if (!no_files) {
            rv = processInputFile(arg);
        } else if (PMAP_FIND_TYPE_PROTOPORT == pmap_type) {
            rv = processProtoPort(arg, NULL, 0);
        } else {
            rv = processAddress(arg, NULL, 0);
        }
    }

    /* done */
    appTeardown();

    return ((0 == rv) ? EXIT_SUCCESS : EXIT_FAILURE);
}


/*
** Local Variables:
** mode:c
** indent-tabs-mode:nil
** c-basic-offset:4
** End:
*/
